import nico
import util
import strutils

{.this: self.}

type
  Coord = tuple[x,y: int]
  Dir = enum
    Right
    Down
    Left
    Up
  Object = ref object of RootObj
    name: string
    x,y: int
    w,h: int
    dir: Dir
    startPos: Coord
    startDir: Dir
    requirements: seq[proc(self: Object): bool {.nimcall.}]
    errors: seq[string]
    transparent: bool
    canReadThrough: bool
    description: string
    # should be able to have multiple rectangles for L shapes etc
  Move = tuple[obj: Object, startPos: Coord, startDir: Dir]
  Room = ref object of RootObj
    # tilemap reference area
    x,y,w,h: int
    doorCoord: Coord
    playerCoord: Coord
    objects: seq[Object]
    completed: bool
    history: seq[Move]
    bestScore: int
    locked: bool

proc inc(x: var Dir) =
  var xi = (x.int + 1)
  if xi > Dir.high.int:
    xi = Dir.low.int
  x = xi.Dir

proc dec(x: var Dir) =
  var xi = (x.int - 1)
  if xi < Dir.low.int:
    xi = Dir.high.int
  x = xi.Dir

proc opposite(x: Dir): Dir =
  result =
    case x:
    of Left: Right
    of Right: Left
    of Up: Down
    of Down: Up

# globals
var room: Room
var cx,cy: int
var heldObject: Object
var showBad: int
var level = 0
var helpTimer = 0
var menuStatic = 60
var camerax,cameray: float
var unlockedLevels = 0
var screenshake: int


method draw(self: Object) {.base.} =
  rectfill(x * 8, y * 8, x * 8 + w * 8 - 1, y * 8 + h * 8 - 1)

const powerPoints = [
  114'u8,
  101,
  195,
  184,
  197,
  146,
  152,
]

const windows = [
  168'u8,
  196,
  197,
  146,
  162,
  152,
]

const niceThings = [
  "looks good",
  "great",
  "cool",
  "that'll do",
  "thanks",
  "not bad",
  "ok",
]

const successMessages = [
  "wow, moving is\ntiring work",
  "thanks so much, mate!",
  "i'll totally help\nyou move next time",
  "thanks, come in\nfor a cuppa",
  "thanks, i owe\nyou a beer!",
]

const curses = [
  "oi!",
  "stop that!",
  "let me go!",
  "move furniture not me!",
  "argh!",
]

var message: string

const sfxPickup = 0
const sfxDrop = 1
const sfxBump = 2
const sfxFlip = 3
const sfxDeny = 4
const sfxBlip = 5
const sfxChangeLevel = 6


proc getObjectAtCell(x,y:int): Object

proc getAdjacentOpenTiles(c: Coord): seq[Coord] =
  result = newSeq[Coord]()
  let x = c.x
  let y = c.y
  try:
    if mget(x-1,y) == 115:
      if getObjectAtCell(x,y) == nil:
        result.add((x-1,y))
  except IndexError:
    discard
  try:
    if mget(x+1,y) == 115:
      if getObjectAtCell(x,y) == nil:
        result.add((x+1,y))
  except IndexError:
    discard
  try:
    if mget(x,y-1) == 115:
      if getObjectAtCell(x,y) == nil:
        result.add((x,y-1))
  except IndexError:
    discard
  try:
    if mget(x,y+1) == 115:
      if getObjectAtCell(x,y) == nil:
        result.add((x,y+1))
  except IndexError:
    discard


proc getAdjacentTiles(self: Object): seq[Coord] =
  result = newSeq[Coord]()
  for dy in y..<y+h:
    for dx in x..<x+w:
      if dx == x and dx > 0:
        result.add((x:dx-1,y:dy))
      if dx == x+w-1 and dx < mapWidth - 1:
        result.add((x:dx+1,y:dy))
      if dy == y and dy > 0:
        result.add((x:dx,y:dy-1))
      if dy == y+h-1 and dy < mapHeight - 1:
        result.add((x:dx,y:dy+1))

proc getAdjacentOnFacing(self: Object): seq[tuple[x,y: int]] =
  # returns adjacent tiles only in the direction the object is facing
  result = newSeq[tuple[x,y: int]]()
  for dy in y..<y+h:
    for dx in x..<x+w:
      if dx == x and dir == Left:
        result.add((x:dx-1,y:dy))
      if dx == x+w-1 and dir == Right:
        result.add((x:dx+1,y:dy))
      if dy == y and dir == Up:
        result.add((x:dx,y:dy-1))
      if dy == y+h-1 and dir == Down:
        result.add((x:dx,y:dy+1))

proc isOccupied(x,y: int): bool =
  # returns true if there is an object or blocking tile there
  if mget(x,y) != 115:
    return true
  for o in room.objects:
    if x >= o.x and x <= o.x + o.w - 1 and y >= o.y and y <= o.y + o.h - 1:
      return true
  return false

proc objectAtCoord(x,y: int): Object =
  # returns object at location
  for o in room.objects:
    if x >= o.x and x <= o.x + o.w - 1 and y >= o.y and y <= o.y + o.h - 1:
      return o
  return nil

proc isReadable(x,y: int): bool =
  # returns true if this tile can be read through
  if mget(x,y) != 115:
    return false
  for o in room.objects:
    if x >= o.x and x <= o.x + o.w - 1 and y >= o.y and y <= o.y + o.h - 1:
      if not o.canReadThrough:
        return false
  return true


method isBlocked(self: Object, x,y: int): bool {.base.} =
  for x in x..<x+w:
    for y in y..<y+h:
      let t = mget(x,y)
      if not (t in [115'u8,116,100]):
        return true
      else:
        discard
  for o in room.objects:
    if o != self:
      if x + w - 1 >= o.x and x <= o.x + o.w - 1 and y + h - 1 >= o.y and y <= o.y + o.h - 1:
        return true
  return false


# requirements

proc isInRoom(self: Object): bool =
  for x in x..<x+w:
    for y in y..<y+h:
      if mget(x,y) != 115:
        return false
  return true

proc isNotInRoom(self: Object): bool =
  for x in x..<x+w:
    for y in y..<y+h:
      if mget(x,y) == 115:
        return false
  return true

proc isAccessible(coord: Coord): bool {.locks:0.} =
  # checks if it can be reached from the doorway
  var queue = newSeq[Coord]()
  queue.add(room.doorCoord)
  var seen = newSeq[Coord]()

  while queue.len > 0:
    var v = queue.pop()
    if v == coord:
      return true

    seen.add(v)
    for adj in getAdjacentOpenTiles(v):
      if not (adj in seen):
        queue.add(adj)

  return false


proc isAccessible(self: Object): bool {.locks:0.} =
  # checks if it can be reached from the doorway
  var queue = newSeq[Coord]()
  queue.add(room.doorCoord)
  var seen = newSeq[Coord]()

  while queue.len > 0:
    var v = queue.pop()
    if getObjectAtCell(v.x,v.y) == self:
      return true

    seen.add(v)
    for adj in getAdjacentOpenTiles(v):
      if not (adj in seen):
        queue.add(adj)

  return false

proc isReadable(self: Object): bool =
  # checks if it is facing blank space
  for coord in getAdjacentOnFacing():
    if not isReadable(coord.x,coord.y):
      return false
    elif not isAccessible(coord):
      return false
  return true

proc isFacingEmptySpace(self: Object): bool =
  # checks if it is facing blank space
  for coord in getAdjacentOnFacing():
    if isOccupied(coord.x,coord.y):
      return false
  return true

proc hasPower(self: Object): bool =
  for coord in getAdjacentTiles():
    if mget(coord.x,coord.y) in powerPoints:
      return true
  return false

proc hasSunlight(self: Object): bool =
  for coord in getAdjacentTiles():
    if mget(coord.x,coord.y) in windows:
      return true
  return false

proc notBlockingWindow(self: Object): bool =
  for coord in getAdjacentTiles():
    if mget(coord.x,coord.y) in windows:
      return false
  return true

proc isOK(self: Object): (bool,seq[string]) =
  # returns true if it satisfies all its requirements
  # check if it doesn't overlap other objects and is inside walls
  result[0] = true
  result[1] = newSeq[string]()
  for i, req in requirements:
    if req(self) == false:
      result[0] = false
      result[1].add(errors[i])

type
  Bed = ref object of Object
  Lounge = ref object of Object
  Armchair = ref object of Object
  Bookshelf = ref object of Object
  CoffeeTable = ref object of Object
  Plant = ref object of Object
  Desk = ref object of Object
  Deskchair = ref object of Object

  TV = ref object of Object
  Fridge = ref object of Object
  Lamp = ref object of Object

  Garbage = ref object of Object

  Player = ref object of Object
    startSpr: int

proc adjacentToDesk(self: Object): bool =
  for coord in getAdjacentOnFacing():
    let o = objectAtCoord(coord.x, coord.y)
    if o of Desk and o.dir == self.dir.opposite:
      return true
  return false

proc canSeeTV(self: Object): bool {.locks:0.} =
  # scan in a line from each part of object in dir until we each an obstacle, if it's a TV success
  for y in y..<y+h:
    for x in x..<x+w:
      var rx = x
      var ry = y
      while true:
        case dir:
        of Left:
          rx -= 1
        of Right:
          rx += 1
        of Up:
          ry -= 1
        of Down:
          ry += 1
        if mget(rx,ry) != 115:
          break
        let o = getObjectAtCell(rx,ry)
        if o != nil:
          if o of TV:
            # check TV is facing the right way
            if o.dir == self.dir.opposite:
              return true
          elif not o.transparent:
            break
  return false

proc newPlayer(x,y: int, startSpr: int): Player =
  result = new(Player)
  result.name = if startSpr == 1: "jez" else: "ivy"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.startSpr = startSpr

  result.requirements = @[]
  result.errors = @[]

  result.description = "keeps asking you\nto help them move\nthat's what friends are for?"

proc newBed(x,y: int, dir: Dir): Bed =
  result = new(Bed)
  result.name = "bed"
  result.x = x
  result.y = y
  result.w = 2
  result.h = 1
  result.dir = dir
  result.requirements = @[
    isInRoom,
    isAccessible,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto bed",
  ]

  result.description = "requires access"

proc newCoffeeTable(x,y: int, dir: Dir): CoffeeTable =
  result = new(CoffeeTable)
  result.name = "coffee table"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir

  result.transparent = true

  result.requirements = @[
    isInRoom,
    isAccessible,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the coffee table",
  ]

  result.description = "you can watch tv\nover a coffee table\nneeds access"

proc newLounge(x,y: int, dir: Dir): Lounge =
  result = new(Lounge)
  result.name = "lounge"
  result.x = x
  result.y = y
  result.w = 2
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    isAccessible,
    isFacingEmptySpace,
    canSeeTV,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the lounge",
    "i can't sit\non the lounge",
    "i can't see\nthe tv from the lounge",
  ]

  result.description = "for watching tv\ntv must be visible\nneeds space for feet"

proc newArmchair(x,y: int, dir: Dir): Armchair =
  result = new(Armchair)
  result.name = "armchair"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    isAccessible,
    isFacingEmptySpace,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the armchair",
    "i can't sit\non the armchair",
  ]

  result.description = "just for relaxing\nneeds access\nand space for feet"


proc newPlant(x,y: int, dir: Dir): Plant =
  result = new(Plant)
  result.name = "plant"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir
  result.requirements = @[
    isInRoom,
    hasSunlight,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "the plant needs\nsunlight",
  ]

  result.description = "just needs sunlight\ndoesn't seem to need water\nno access required"

proc newBookshelf(x,y: int, dir: Dir): Bookshelf =
  result = new(Bookshelf)
  result.name = "bookshelf"
  result.x = x
  result.y = y
  result.w = 2
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    isAccessible,
    isReadable,
    notBlockingWindow,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the bookshelf",
    "i can't see\nthe books",
    "the bookshelf is\nblocking the window",
  ]

  result.description = "blocks windows\nbooks must be accessible"


proc newFridge(x,y: int, dir: Dir): Fridge =
  result = new(Fridge)
  result.name = "fridge"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    isAccessible,
    isReadable,
    hasPower,
    notBlockingWindow,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the fridge",
    "i can't open\nthe fridge",
    "the fridge\nneeds power",
    "the fridge is\nblocking the window",
  ]

  result.description = "blocks windows\nrequires power\nand space to open"

proc newTV(x,y: int, dir: Dir): TV =
  result = new(TV)
  result.name = "tv"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    isAccessible,
    isReadable,
    hasPower,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the tv",
    "i can't see\nthe screen",
    "the tv needs\npower",
  ]

  result.description = "requires power\nmust be accessible\nlost the remote"

proc newLamp(x,y: int, dir: Dir): Lamp =
  result = new(Lamp)
  result.name = "lamp"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    hasPower,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "the lamp\nneeds power",
  ]

  result.description = "requires power\nactivated by clapping\nno access required"

proc newDeskchair(x,y: int, dir: Dir): Deskchair =
  result = new(Deskchair)
  result.name = "deskchair"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir
  result.canReadThrough = true

  result.requirements = @[
    isInRoom,
    isAccessible,
    adjacentToDesk,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get to\nthe chair",
    "needs to go\nnext to desk",
  ]

  result.description = "used at a desk"

proc newDesk(x,y: int, dir: Dir): Desk =
  result = new(Desk)
  result.name = "desk"
  result.x = x
  result.y = y
  result.w = 2
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isInRoom,
    isAccessible,
    isReadable,
    hasPower,
  ]
  result.errors = @[
    "that needs to\ngo inside",
    "i can't get\nto the desk",
    "i can't use\nthe desk",
    "the desk needs\npower",
  ]

  result.description = "requires power\nneeds access to work"

proc newGarbage(x,y: int, dir: Dir): Garbage =
  result = new(Garbage)
  result.name = "garbage"
  result.x = x
  result.y = y
  result.w = 1
  result.h = 1
  result.dir = dir

  result.requirements = @[
    isNotInRoom,
  ]
  result.errors = @[
    "take out the\ngarbage",
  ]

  result.description = "must be removed from the room"

method rotate(self: Object, onCollide: proc() = nil) {.base.} =
  let oldDir = dir
  dir.inc()
  swap(self.w,self.h)
  # check if it would be blocked
  if isBlocked(x,y):
    swap(self.w,self.h)
    dir = oldDir
    if onCollide != nil:
      onCollide()

method setDir(self: Object, newDir: Dir) {.base.} =
  case dir:
  of Left,Right:
    case newDir:
    of Up,Down:
      swap(self.w,self.h)
    else:
      discard
  of Up,Down:
    case newDir:
    of Left,Right:
      swap(self.w,self.h)
    else:
      discard
  dir = newDir

method draw(self: Bed) =
  case dir:
  of Left:
    spr(32,x*8,y*8,2,1,true)
  of Right:
    spr(32,x*8,y*8,2,1)
  of Up:
    spr(40,x*8,y*8,1,2)
  of Down:
    spr(34,x*8,y*8,1,2)

method draw(self: Lounge) =
  case dir:
  of Left:
    spr(72,x*8,y*8,1,2,true)
  of Right:
    spr(72,x*8,y*8,1,2)
  of Up:
    spr(70,x*8,y*8,2,1)
  of Down:
    spr(48,x*8,y*8,2,1)

method draw(self: Armchair) =
  case dir:
  of Left:
    spr(64,x*8,y*8,1,1,true)
  of Right:
    spr(64,x*8,y*8,1,1)
  of Up:
    spr(82,x*8,y*8,1,1)
  of Down:
    spr(65,x*8,y*8,1,1)

method draw(self: Bookshelf) =
  case dir:
  of Left:
    spr(35,x*8,y*8,1,2)
  of Right:
    spr(35,x*8,y*8,1,2,true)
  of Up:
    spr(52,x*8,y*8,2,1)
  of Down:
    spr(36,x*8,y*8,2,1)

method draw(self: Desk) =
  case dir:
  of Left:
    spr(73,x*8,y*8,1,2,true)
  of Right:
    spr(73,x*8,y*8,1,2)
  of Up:
    spr(86,x*8,y*8,2,1)
  of Down:
    spr(68,x*8,y*8,2,1)

method draw(self: CoffeeTable) =
  spr(66,x*8,y*8,1,1)

method draw(self: Fridge) =
  case dir:
  of Left:
    spr(55,x*8,y*8,1,1)
  of Right:
    spr(55,x*8,y*8,1,1, true)
  of Up:
    spr(42,x*8,y*8,1,1)
  of Down:
    spr(54,x*8,y*8,1,1)

method draw(self: TV) =
  case dir:
  of Left:
    spr(81,x*8,y*8,1,1, true)
  of Right:
    spr(81,x*8,y*8,1,1)
  of Up:
    spr(96,x*8,y*8,1,1)
  of Down:
    spr(80,x*8,y*8,1,1)

method draw(self: Deskchair) =
  case dir:
  of Left:
    spr(85,x*8,y*8,1,1, true)
  of Right:
    spr(85,x*8,y*8,1,1)
  of Up:
    spr(112,x*8,y*8,1,1)
  of Down:
    spr(84,x*8,y*8,1,1)

method draw(self: Plant) =
  spr(67,x*8,y*8,1,1)

method draw(self: Lamp) =
  spr(83,x*8,y*8,1,1)

method draw(self: Garbage) =
  spr(41,x*8,y*8,1,1)

method draw(self: Player) =
  spr(startSpr,x*8,y*8)

proc newRoom(x,y,w,h: int): Room =
  result = new(Room)
  result.x = x
  result.y = y
  result.w = w
  result.h = h
  result.completed = false
  result.objects = newSeq[Object]()
  result.history = newSeq[Move]()

  # go through the map and extract any objects
  for y in y..<y+h:
    for x in x..<x+w:
      let t = mget(x,y)
      if t == 1 or t == 17:
        result.playerCoord = (x,y)
        result.objects.add(newPlayer(x,y,t.int))
        mset(x,y,209)
      elif t == 100:
        result.doorCoord = (x,y)
      elif t == 32:
        mset(x,y,116)
        mset(x+1,y,116)
        result.objects.add(newBed(x,y,Left))
      elif t == 48:
        mset(x,y,116)
        mset(x+1,y,116)
        result.objects.add(newLounge(x,y,Down))
      elif t == 65:
        mset(x,y,116)
        result.objects.add(newArmchair(x,y,Down))
      elif t == 66:
        mset(x,y,116)
        result.objects.add(newCoffeeTable(x,y,Left))
      elif t == 67:
        mset(x,y,116)
        result.objects.add(newPlant(x,y,Left))
      elif t == 36:
        mset(x,y,116)
        mset(x+1,y,116)
        result.objects.add(newBookshelf(x,y,Down))
      elif t == 54:
        mset(x,y,116)
        result.objects.add(newFridge(x,y,Down))
      elif t == 68:
        mset(x,y,116)
        mset(x+1,y,116)
        result.objects.add(newDesk(x,y,Down))
      elif t == 80:
        mset(x,y,116)
        result.objects.add(newTV(x,y,Down))
      elif t == 84:
        mset(x,y,116)
        result.objects.add(newDeskchair(x,y,Down))
      elif t == 83:
        mset(x,y,116)
        result.objects.add(newLamp(x,y,Down))
      elif t == 41:
        mset(x,y,115)
        result.objects.add(newGarbage(x,y,Down))

proc getUnlockedLevels() =
  var unlocked = 0.0
  for i in 0..15:
    try:
      if parseInt(getConfigValue("score", "level_" & $i)) > 0:
        unlocked += 1.25
    except ValueError:
      discard
  unlockedLevels = unlocked.int

proc getCompletedLevels(): int =
  for i in 0..15:
    try:
      if parseInt(getConfigValue("score", "level_" & $i)) > 0:
        result += 1
    except ValueError:
      discard

proc loadLevel(level: int) =
  let x = (level mod 8) * 16
  let y = (level div 8) * 12

  loadMapFromJson("ld37.json")
  room = newRoom(x, y, 16, 12)
  heldObject = nil
  if room.objects.len > 0:
    cx = room.objects[0].x
    cy = room.objects[0].y
  else:
    cx = room.x
    cy = room.y

  if level == 0:
    cx = 8
    cy = 8

  try:
    room.bestScore = parseInt(getConfigValue("score", "level_" & $level))
  except ValueError:
    room.bestScore = 0

  getUnlockedLevels()
  room.locked = level > unlockedLevels

  message = nil
  helpTimer = 120


proc draw(self: Room) =
  # draw floor

  setColor(0)
  rectfill(x*8,y*8,x*8+w*8,y*8+h*8)

  mapDraw(x,y,w,h,x*8,y*8)

  for o in objects:
    o.draw()

###

proc menuUpdate(dt: float)
proc menuDraw()
proc menuInit() =
  frame = 0
  sfx(sfxChangeLevel)
  music(0)
  menuStatic = 30

### GAME MODE

proc gameInit() =
  frame = 0
  music(0)

proc getObjectAtCell(x,y:int): Object =
  for o in room.objects:
    if x >= o.x and x <= o.x + o.w - 1 and y >= o.y and y <= o.y + o.h - 1:
      return o
  return nil

method move(self: Object, dir: Dir, onCollide: proc() = nil) {.base.} =
  case dir:
  of Up:
    if not isBlocked(x,y-1):
      y -= 1
    elif onCollide != nil:
      onCollide()
  of Down:
    if not isBlocked(x,y+1):
      y += 1
    elif onCollide != nil:
      onCollide()
  of Left:
    if not isBlocked(x-1,y):
      x -= 1
    elif onCollide != nil:
      onCollide()
  of Right:
    if not isBlocked(x+1,y):
      x += 1
    elif onCollide != nil:
      onCollide()

method move(self: Player, dir: Dir, onCollide: proc() = nil) =
  message = rnd(curses)
  helpTimer = 120
  if onCollide != nil:
    onCollide()

proc startMainMusic() {.cdecl.} =
  onMusicFinished(nil)
  music(0)

proc gameUpdate(dt: float) =
  if btnp(pcStart) or btnp(pcBack):
    nico.run(menuInit, menuUpdate, menuDraw)

  if room.completed:
    if btnp(pcA) or btnp(pcX):
      level += 1
      if level > 15:
        level = 15
      loadLevel(level)
      nico.run(menuInit, menuUpdate, menuDraw)

  else:

    if btnp(pcLeft):
      cx -= 1
      if heldObject != nil:
        heldObject.move(Left) do():
          cx += 1
          sfx(sfxBump)
          screenShake += 5
    elif btnp(pcRight):
      cx += 1
      if heldObject != nil:
        heldObject.move(Right) do():
          cx -= 1
          sfx(sfxBump)
          screenShake += 5
    elif btnp(pcUp):
      cy -= 1
      if heldObject != nil:
        heldObject.move(Up) do():
          cy += 1
          sfx(sfxBump)
          screenShake += 5
    elif btnp(pcDown):
      cy += 1
      if heldObject != nil:
        heldObject.move(Down) do():
          cy -= 1
          sfx(sfxBump)
          screenShake += 5

    elif btnp(pcA):
      # select / deselect object
      if heldObject != nil:
        if heldObject.startPos != (heldObject.x,heldObject.y) or heldObject.startDir != heldObject.dir:
          room.history.add((
            obj: heldObject,
            startPos: heldObject.startPos,
            startDir: heldObject.startDir,
          ))
        if heldObject.isOK[0]:
          if not (heldObject of Player):
            message = rnd(niceThings)
            helpTimer = 60
        heldObject = nil
        # drop
        sfx(sfxDrop)
        screenShake += 2
      else:
        let o = getObjectAtCell(cx,cy)
        if o != nil:
          heldObject = o
          heldObject.startPos = (heldObject.x, heldObject.y)
          heldObject.startDir = heldObject.dir
          sfx(sfxPickup)
        else:
          # tried to pick up nothing, show bad objects
          showBad = 60

    elif btnp(pcX):
      if heldObject != nil:
        # rotate object
        heldObject.rotate() do():
          sfx(sfxBump)
          screenShake += 5
        sfx(sfxFlip)

    elif btnp(pcY):
      if heldObject == nil:
        # undo history
        if room.history.len > 0:
          let lastMove = room.history.pop()
          lastMove.obj.x = lastMove.startPos.x
          lastMove.obj.y = lastMove.startPos.y
          lastMove.obj.setDir(lastMove.startDir)
          sfx(sfxChangeLevel)
          cx = lastMove.obj.x
          cy = lastMove.obj.y

    if heldObject == nil:
      var completed = true
      for o in room.objects:
        if not o.isOK[0]:
          completed = false
          break
      if completed:
        room.completed = true
        if room.bestScore == 0 or room.history.len < room.bestScore:
          room.bestScore = room.history.len
          updateConfigValue("score", "level_" & $level, $room.bestScore)
          saveConfig()
        message = rnd(successMessages)
        helpTimer = 240
        music(-1)
        music(1,0)
        onMusicFinished(startMainMusic)

  if message == nil:
    if helpTimer > 0:
      helpTimer -= 1
    else:
      if heldObject != nil:
        let (ok,issues) = heldObject.isOK()
        if not ok:
          setColor(8)
          for issue in issues:
            message = issue
            helpTimer = 120
            break
      else:
        # find an object (in the house) with a problem and write a message
        if room.objects.len > 0:
          let obj = rnd(room.objects)
          if obj.isInRoom():
            let (ok,issues) = obj.isOK()
            if not ok:
              message = rnd(issues)
              helpTimer = 120
  else:
    helpTimer -= 1
    if helpTimer == 0:
      message = nil
      helpTimer = 120


proc drawBubble(text: string, x,y: int, yell: bool = false) =
  let w = getMaxLineLength(text) * 4 + 16
  let x = x - w div 2
  var sprStart = 0
  if yell:
    sprStart += 32

  spr(sprStart+13,x,y)
  spr(sprStart+15,x+w-8,y)
  spr(sprStart+29,x,y+8)
  spr(sprStart+31,x+w-8,y+8)

  setColor(if yell: 15 else: 7)
  rectfill(x+8,y,x+w-8,y+15)
  setColor(13)
  line(x+8, y-1, x+w-8, y-1)
  line(x+8, y+16, x+w-8, y+16)

  let sx = room.playerCoord.x * 8 + 4 - room.x * 8
  let sy = room.playerCoord.y * 8 + 8 - room.y * 8

  setColor(if yell: 15 else: 7)
  line(x + w div 2 - 1, y - 1, sx, sy)
  line(x + w div 2 + 1, y - 1, sx, sy)
  line(x + w div 2 - 2, y - 1, sx, sy)
  line(x + w div 2 + 2, y - 1, sx, sy)
  setColor(13)
  line(x + w div 2 - 3, y - 1, sx, sy)
  line(x + w div 2 + 3, y - 1, sx, sy)
  setColor(if yell: 15 else: 7)
  line(x + w div 2, y - 1, sx, sy)

  setColor(1)
  printLinesC(text, x + w div 2, y + 2 + (if text.getLines() == 0: 3 else: 0))


proc drawInfo(self: Object) =
  setColor(9)
  print(name, 2, 92)
  setColor(6)
  printLines(description, 2, 92+8)

proc gameDraw() =
  cls()

  setCamera(room.x*8,room.y*8)
  if screenshake > 0:
    setCamera(room.x * 8 + rnd(2)-1, room.y * 8 + rnd(2)-1)

  room.draw()

  if showBad > 0:
    if frame mod 30 < 15:
      setColor(8)
      for o in room.objects:
        if not o.isOK[0]:
          rect(o.x*8-1, o.y*8-1, o.x*8+o.w*8, o.y*8+o.h*8)
      showBad -= 1

  if not room.completed:
    if heldObject != nil:
      # highlight current object
      if frame mod 30 < 20:
        if not heldObject.isOK[0]:
          setColor(8)
        else:
          setColor(9)
        rect(heldObject.x*8-1, heldObject.y*8-1, heldObject.x*8+heldObject.w*8, heldObject.y*8+heldObject.h*8)

    else:
      # draw cursor
      if frame mod 30 < 20:
        pal(9,10)
      spr(11, cx*8 + 3, cy*8 + 6)
      pal()

  setCamera()

  if screenshake > 0:
    screenshake -= 1
    setCamera(rnd(2)-1,rnd(2)-1)

  # draw gui
  setColor(1)
  rectfill(0,90,127,127)
  setColor(6)
  line(0,89,127,89)

  setColor(6)
  printr("moves: " & $room.history.len, 127, 92)
  if room.bestScore > 0:
    setColor(9)
    printr("best:  " & $room.bestScore, 127, 92+8)
  if room.completed:
    setColor(if frame mod 30 < 15: 9 else: 7)
    printShadowC("moved in!", 64, 42, 2)
    printShadowC($room.history.len & " moves", 64, 42 + 16)

  if not room.completed:
    if heldObject != nil:
      heldObject.drawInfo()
    else:
      let o = getObjectAtCell(cx,cy)
      if o != nil:
        o.drawInfo()
      else:
        setColor(9)
        print("how to play", 2, 92)
        setColor(6)
        printLines("move all the furniture\ninto the carpeted room\nuse arrows, z, x and c", 2, 92+8)

  if message != nil:
    drawBubble(message, 64, 92, '!' in message)

  # controls
  setColor(6)
  if not room.completed:
    if heldObject == nil:
      let o = getObjectAtCell(cx,cy)
      if o != nil:
        richPrint("z: <9>pickup</> c:   <9>undo", 127, 121, taRight)
      else:
        richPrint("z: <9>check</>  c:   <9>undo", 127, 121, taRight)
    else:
      richPrint(  "z: <9>place</>  x: <9>rotate", 127, 121, taRight)
  else:
    richPrint("z or x: <9>menu", 127, 121, taRight)

### MENU MODE

proc mainInit() =
  setWindowTitle("Moving In - by Impbox for LD37")
  setTargetSize(128,128)
  setScreenSize(256,256)
  loadSpritesheet("spritesheet.png")
  loadMapFromJson("ld37.json")

  loadSfx(sfxPickup, "sfx/pickup.ogg")
  loadSfx(sfxDrop, "sfx/drop.ogg")
  loadSfx(sfxBump, "sfx/bump.ogg")
  loadSfx(sfxFlip, "sfx/flip.ogg")
  loadSfx(sfxBlip, "sfx/blip.ogg")
  loadSfx(sfxChangeLevel, "sfx/changelevel.ogg")
  loadSfx(sfxDeny, "sfx/deny.ogg")

  loadMusic(0, "music/ld37music.ogg")
  loadMusic(1, "music/success.ogg")
  musicVol(64)

  loadConfig()
  loadLevel(0)

  screenshake = 0

proc menuUpdate(dt: float) =
  if btnp(pcBack):
    shutdown()
    return
  elif btnp(pcLeft):
    level -= 1
    if level < 0:
      level = 0
      sfx(sfxBump)
    else:
      sfx(sfxChangeLevel)
      loadLevel(level)
  elif btnp(pcRight):
    level += 1
    if level > 15:
      level = 15
      sfx(sfxBump)
    else:
      loadLevel(level)
      sfx(sfxChangeLevel)
  elif btnp(pcA):
    if not room.locked:
      nico.run(gameInit, gameUpdate, gameDraw)
      music(0)
      sfx(sfxChangeLevel)
    else:
      sfx(sfxDeny)
      screenshake = 15

  if menuStatic > 0:
    menuStatic -= 1

  if menuStatic <= 0:
    camerax = lerp(camerax, room.x.float * 8.0, 0.1)
    cameray = lerp(cameray, room.y.float * 8.0, 0.1)
    if abs(camerax - room.x.float * 8.0) < 0.5:
      camerax = room.x.float * 8.0
    if abs(cameray - room.y.float * 8.0) < 0.5:
      cameray = room.y.float * 8.0

proc menuDraw() =
  # draw level preview
  setCamera(camerax.int,cameray.int)
  room.draw()
  setCamera()

  if screenshake > 0:
    setCamera(rnd(4)-2,rnd(4)-2)
    screenshake -= 1

  if menuStatic > 0 or room.locked:
    for y in 0..<screenHeight:
      for x in 0..<screenWidth:
        pset(x,y,rnd([0,1,6,7]))

  setColor(1)
  rectfill(0,0,127,7)
  setColor(6)
  line(0,7,127,7)

  setColor(9)
  print("level " & $(level+1), 1, 1)
  if room.locked:
    setColor(8)
    printr("locked", 128, 1)
  elif room.bestScore > 0:
    printr("completed in " & $room.bestScore & " moves", 128, 1)
  else:
    printr("not completed", 128, 1)

  setColor(1)
  rectfill(0,90,127,127)
  setColor(6)
  line(0,89,127,89)

  # logo
  sspr(32,0,50,11, 2, 92, 100, 22)


  richPrint("<- -> : <9>select level</> z: <9>begin", 127, 121, taRight)

  if frame < 5:
    setColor(7)
    rectfill(0,0,screenWidth,screenHeight)

proc introUpdate(dt: float) =
  if btnp(pcStart) or btnp(pcA):
    nico.run(menuInit, menuUpdate, menuDraw)

var drippiness = 1000

proc introDraw() =
  if frame == 60:
    cls()
    sfx(sfxDrop)
    sspr(80,24, 24,24, 64 - 12, 64 - 12, 24, 24)
  elif frame == 120:
    sfx(sfxDrop)
    setColor(7)
    printShadowC("ld37", 64, 92)
  elif drippiness > 0:
    # do drippy effect
    for i in 0..<drippiness:
      let x = rnd(128)
      let y = rnd(128)
      let c = pget(x,y)
      if c == 12:
        if pget(x,y+1) != 7:
          pset(x,y+1,c)
          drippiness -= 1

  if frame >= 300:
    nico.run(menuInit, menuUpdate, menuDraw)






nico.init()
nico.run(mainInit, introUpdate, introDraw)
